/* Prefetch Abort handling
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Image Loader.

The Imote2 Image Loader is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

The Imote2 Image Loader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the Imote2 Image Loader. If not, see <http://www.gnu.org/licenses/>.
*/
#include "exception.h"
#include "exceptions.h"
#include "ledcycle.h"
#include "doublefault.h"
#include "backtrace.h"
#include <stdio.h>

static void printFault(uint32_t lr, const uint32_t * fp, uint32_t fault,
	const struct _ExcProcessorStatus * pstat);

__attribute__((interrupt("ABORT"),weak))
void hplarmv_pabort() {
	/* gather some data: lr, fp, fault source, fault address and processor status */
	uint32_t lr = (uint32_t)__builtin_return_address(0);
	uint32_t * fp = (uint32_t*)__builtin_frame_address(0);
	uint32_t fault = _excGetFaultStatus();
	struct _ExcProcessorStatus pstat;
	_excGetProcessorStatus(&pstat);
	
	/* check for double faults */
	_excDoubleFaultCheck(lr, fp, &pstat);
	
	/* initialize exception handler */
	_excInit();
	
	/* begin in next line */
	_excUartOut("\n");
	
	while (1) {
		printFault(lr, fp, fault, &pstat);
		_excLedCycle(TRUE);
	}
}

static void printFault(uint32_t lr, const uint32_t * fp, uint32_t fault,
	const struct _ExcProcessorStatus * pstat) {
	char buf[128];
	
	/* print the exception type and the old mode */
	snprintf(buf, sizeof(buf), "PREFETCH ABORT occured in %s mode\n", pstat->spsrModeLabel);
	_excUartOut(buf);
	
	/* now get the abortion source and fault address */
	uint32_t src = ((fault >> 6) & 0x10) | (fault & 0xf);
	const char * sourceLabel = NULL;
	switch (src) {
		case 0x10:
			/* Instruction MMU Exception */
			sourceLabel = "Instruction MMU Exception";
			break;
		case 0x16:
			/* External Instruction Error Exception */
			sourceLabel = "External Instruction Error";
			break;
		case 0x18:
			/* Instruction Cache Parity Error Exception */
			sourceLabel = "Instruction Cache Parity Error";
			break;
		default:
			sourceLabel = "Unknown Abort";
			break;
	}
	/* print the pretty printed fault status */
	snprintf(buf, sizeof(buf), " Fault Status: %s\n", sourceLabel);
	_excUartOut(buf);
	/* and some more information */
	snprintf(buf, sizeof(buf), " FSR:  0x%08x  FPC: 0x%08x        SP: 0x%08x\n", fault,
		lr - 4, pstat->oldsp);
	_excUartOut(buf);
	
	_excPrintProcessorStatus(pstat->oldlr, fp, pstat);
	
	_excBacktrace(pstat->oldlr, fp);
}

